/*
 * Copyright 2023 University of Liège
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

// SWIG input file for python interface

%feature("autodoc","1");

%module(docstring=
"'sdpmw' module: Source-Doublet Panel Method
(c) ULg - A&M",
threads="1"
) sdpmw
%{
#include "sdpm.h"
#include "sdpmObject.h"
#include "sdpmTimers.h"
#include "sdpmMesh.h"
#include "sdpmNode.h"
#include "sdpmElement.h"
#include "sdpmLine2.h"
#include "sdpmQuad4.h"
#include "sdpmTag.h"
#include "sdpmGroup.h"
#include "sdpmGmshImport.h"
#include "sdpmGmshExport.h"
#include "sdpmUnitTest.h"

#include "sdpmBody.h"
#include "sdpmProblem.h"
#include "sdpmSolver.h"
#include "sdpmSolverTransonic.h"
#include "sdpmAdjoint.h"
#include "sdpmGradient.h"

#include "sdpmCppBuf2Py.h"

#include <sstream>
%}

// COMMON
%include "std_string.i"
%include "std_vector.i"
%include "std_map.i"
%include "std_complex.i"

// EXCEPTION
// from: http://swig.10945.n7.nabble.com/Trapping-Swig-DirectorException-td6013.html
// le code suivant permet de voir la call stack dans les appels C++ => python
%include "exception.i"
%{
   static void handle_exception(void) {
     try {
       throw;
     } catch (std::exception &e) {
        std::stringstream txt;
        txt << e.what(); // << ' ' << typeid(e).name();
        PyErr_SetString(PyExc_RuntimeError, e.what());
     }
     catch(...)
     {
        PyErr_SetString(PyExc_RuntimeError, "Unknown C++ Runtime Error");
     }
   }
%}
%exception {
   try {
     $action
   } catch (...) {
     // Note that if a director method failed, the Python error indicator
     // already contains full details of the exception, and it will be
     // reraised when we go to SWIG_fail; so no need to convert the C++
     // exception back to a Python one
     if (!PyErr_Occurred()) {
       handle_exception();
     }
     SWIG_fail;
   }
}

// SDPM typemaps
%include "sdpm_def.h" // needed to access USE_CODI in SWIG interface
#ifndef USE_CODI
typedef double sdpmDouble; // needed so that SWIG does not convert sdpmDouble to double
#endif
%typemap(out) sdpmDouble {
#ifndef USE_CODI
    $result = PyFloat_FromDouble($1);
#else
    $result = PyFloat_FromDouble(($1).getValue());
#endif
}
// SDPM templates
%template(std_vector_double) std::vector<double>;
%template(std_vector_vector_double) std::vector<std::vector<double>>;
%template(std_vector_string) std::vector<std::string>;
%template(std_map_string_vector_double) std::map<std::string, std::vector<double>>;
%template(std_vector_Nodep) std::vector<sdpm::Node *>;
%template(std_vector_Elementp) std::vector<sdpm::Element *>;
%template(std_vector_Tagp) std::vector<sdpm::Tag *>;
%template(std_map_string_Tagp) std::map<std::string, sdpm::Tag *>;
%template(std_vector_vector3d) std::vector<sdpmVector3d>;
#ifndef USE_CODI
%rename(std_vector_double) std::vector<sdpmDouble>;
%rename(std_vector_vector_double) std::vector<std::vector<sdpmDouble>>;
#else
%template(std_vector_sdpmdouble) std::vector<sdpmDouble>;
%template(std_vector_vector_sdpmdouble) std::vector<std::vector<sdpmDouble>>;
#endif

// SDPM
%include "sdpm.h"
%include "sdpmObject.h"
%include "sdpmTimers.h"
%include "sdpmMesh.h"
%include "sdpmNode.h"
%nodefault sdpm::Element; // Element is pure virtual
%include "sdpmElement.h"
%include "sdpmLine2.h"
%include "sdpmQuad4.h"
%include "sdpmTag.h"
%include "sdpmGroup.h"
%include "sdpmGmshImport.h"
%include "sdpmGmshExport.h"
%include "sdpmUnitTest.h"

%include "sdpmBody.h"
%include "sdpmProblem.h"
%include "sdpmSolver.h"
%include "sdpmSolverTransonic.h"
%include "sdpmAdjoint.h"
%include "sdpmGradient.h"

%warnfilter(401); // Nothing known about base class 'std::basic_streambuf< char >'
%include "sdpmCppBuf2Py.h"
%warnfilter(+401);

// Interface to SDPM types
class sdpmVector3d {};
%extend sdpmVector3d {
    double __getitem__(int i)
    {
        if (i > 2)
            throw std::runtime_error("sdpmVector3d: index out of range (size 3)!\n");
#ifndef USE_CODI
        return (*self)(i);
#else
        return (*self)(i).getValue();
#endif
    }
    // use repr instead of str so that iterables also print nicely
    std::string __repr__()
    {
        std::ostringstream out; out << *self;
        return out.str();
    }
}

#ifdef USE_CODI
class sdpmDouble {};
%extend sdpmDouble {
    double __float__()
    {
        return (*self).getValue();
    }
    // use repr instead of str so that iterables also print nicely
    std::string __repr__()
    {
        std::ostringstream out; out << *self;
        return out.str();
    }
}
#endif

// Interface to SDPM classes
namespace sdpm {

// So that print(obj) use << *obj
%extend sdpmObject {
    std::string __str__() {
        std::stringstream str;
        str << *self;
        return str.str();
    }
}

// To get a Timer from Timers[]
%extend Timers {
    Timer &getitem(std::string const &name)
    {
        return (*self)[name];
    }

    %pythoncode {
def __getitem__(self, name):
    return self.getitem(name)
    }
}
};
