# -*- coding: utf-8 -*-

# Copyright 2023 University of Liège
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# AGARD 445.6 wing
# Adrien Crovato

def get():
    from sdpm.utils import build_fpath
    wnc = 20
    cfg = {
        # Model (geometry or mesh)
        'File': build_fpath('models/agard445.geo', __file__, 1), # input Gmsh file
        'Pars': {'nC': wnc, 'nS': 10, 'pC': 1.1, 'pS': 1.0}, # parameters for Gmsh file (optional)
        # Markers
        'Wing': 'wing', # name of the marker defining the lifting body (wing)
        'Te': 'te', # name of the marker defining the trailing edge of the lifting body
        # Freestream
        'Mach': 0.9, # Mach number (optional)
        'AoA': 0.0, # angle of attack (optional)
        'AoS': 0.0, # angle of side slip (optional)
        # Unsteady motion
        'Num_wake_div': wnc, # number of cells per chord length in the wake
        'Frequencies': [0.01, 0.1, 0.3, 0.5], # list of reduced frequencies
        'Modes': build_fpath('models/agard445_modes.csv', __file__, 1), # input file containing the modes (optional)
        'Num_modes': 4, # number of modes
        # Geometry
        'Symmetry': True, # whether only half (symmetric) model is provided
        's_ref': 0.353, # reference surface area
        'c_ref': 0.559, # reference chord length
        'x_ref': 0.0, # reference point for moment computation (x)
        'y_ref': 0.0, # reference point for moment computation (y)
        'z_ref': 0.0 # reference point for moment computation (z)
    }
    return cfg

def main():
    # Init SDPM
    from sdpm.api.core import init_sdpm
    import numpy as np
    sdpm = init_sdpm(get())
    # Run the solver
    sdpm['sol'].update()
    sdpm['sol'].run()
    # Get the GAF matrices
    for i in range(sdpm['n_f']):
        print(f'- GAF at frequency #{i}')
        print(np.array(sdpm['sol'].getGafMatrixRe(i)) + 1j * np.array(sdpm['sol'].getGafMatrixIm(i)))

if __name__ == "__main__":
    main()
