/*
 * Copyright 2023 University of Liège
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include "sdpmAdjoint.h"

#ifdef USE_CODI
#include "sdpmSolver.h"
#include "sdpmProblem.h"
#include "sdpmBody.h"
#include <iostream>

using namespace sdpm;

Adjoint::Adjoint(Solver &sol) : _sol(sol), _nm(_sol._pbl._nModes), _nf(_sol._pbl._omega.size()),
                                _aoa(_sol._pbl._aoa),
                                _omega(_sol._pbl._omega),
                                _cl(_sol._cl), _cd(_sol._cd),
                                _cl1Re(_sol._cl1Re), _cl1Im(_sol._cl1Im),
                                _cd1Re(_sol._cd1Re), _cd1Im(_sol._cd1Im)
{
    // Store coordinates
    for (auto b : _sol._pbl._bodies)
    {
        std::string bName = b->getName();
        std::vector<Node *> nodes = b->getNodes();
        _x[bName].resize(nodes.size());
        for (size_t i = 0; i < nodes.size(); ++i)
            _x[bName][i] = &nodes[i]->_coord;
    }
}

Adjoint::~Adjoint()
{
    std::cout << "~sdpm::Adjoint()\n";
}

/**
 * @brief Run the solver and record the operations
 */
void Adjoint::solve()
{
    // Activate the tape and record the input variables
    std::cout << "Recording operations..." << std::endl;
    sdpmTape &tape = sdpmDouble::getTape();
    tape.reset();
    tape.setActive();
    tape.registerInput(_aoa);
    for (auto b : _x)
        for (auto c : b.second)
            for (int i = 0; i < c->size(); ++i)
                tape.registerInput((*c)(i));
    for (size_t i = 0; i < _nf; ++i)
        tape.registerInput(_omega[i]);
    // Update and solve while recording
    _sol.update();
    _sol.run();
    // Record the output variables and deactivate the tape
    tape.registerOutput(_cl);
    tape.registerOutput(_cd);
    for (size_t i = 0; i < _nm; ++i)
    {
        for (size_t j = 0; j < _nf; ++j)
        {
            tape.registerOutput(_cl1Re[i][j]);
            tape.registerOutput(_cl1Im[i][j]);
            tape.registerOutput(_cd1Re[i][j]);
            tape.registerOutput(_cd1Im[i][j]);
        }
    }
    tape.setPassive();
}

/**
 * @brief Compute the adjoint sensitivities (i.e. perform a reverse evaluation of the tape)
 */
std::map<std::string, std::vector<double>> Adjoint::compute(std::string const &dOutName, std::vector<double> const &dOut, std::vector<std::string> const &dInNames)
{
    // Tape
    sdpmTape &tape = sdpmDouble::getTape();

    // Evaluate the gradients
    std::cout << "Evaluating sensitivities... " << std::flush;
    std::map<std::string, std::vector<double>> dIn;
    tape.clearAdjoints();
    // steady lift coefficient
    if (dOutName == "cl")
        _cl.setGradient(dOut[0]);
    // steady drag coefficient
    else if (dOutName == "cd")
        _cd.setGradient(dOut[0]);
    // real part of harmonic lift coefficient
    else if (dOutName == "cl1_re")
        for (size_t i = 0; i < _nm; ++i)
            for (size_t j = 0; j < _nf; ++j)
                _cl1Re[i][j].setGradient(dOut[i * _nf + j]);
    // imaginary part of harmonic lift coefficient
    else if (dOutName == "cl1_im")
        for (size_t i = 0; i < _nm; ++i)
            for (size_t j = 0; j < _nf; ++j)
                _cl1Im[i][j].setGradient(dOut[i * _nf + j]);
    // real part of harmonic drag coefficient
    else if (dOutName == "cd1_re")
        for (size_t i = 0; i < _nm; ++i)
            for (size_t j = 0; j < _nf; ++j)
                _cd1Re[i][j].setGradient(dOut[i * _nf + j]);
    // imaginary part of harmonic drag coefficient
    else if (dOutName == "cd1_im")
        for (size_t i = 0; i < _nm; ++i)
            for (size_t j = 0; j < _nf; ++j)
                _cd1Im[i][j].setGradient(dOut[i * _nf + j]);
    else
        throw std::runtime_error("Adjoint::compute: wrong output sensitivity variable name provided!\n");
    tape.evaluate();
    std::cout << "done." << std::endl;

    // Store the gradients
    // w.t.r. AoA
    auto it = std::find(dInNames.begin(), dInNames.end(), "aoa");
    if (it != dInNames.end())
        dIn["aoa"].push_back(_aoa.getGradient());
    // w.r.t. coordinates
    for (auto b : _sol._pbl._bodies)
    {
        std::string bName = b->getName();
        it = std::find(dInNames.begin(), dInNames.end(), "x_" + bName);
        if (it != dInNames.end())
            for (size_t i = 0; i < _x[bName].size(); ++i)
                for (int j = 0; j < _x[bName][i]->size(); ++j)
                    dIn["x_" + bName].push_back((*(_x[bName][i]))(j).getGradient());
    }
    // w.r.t. frequencies
    it = std::find(dInNames.begin(), dInNames.end(), "omega");
    if (it != dInNames.end())
        for (size_t i = 0; i < _nf; ++i)
            dIn["omega"].push_back(_omega[i].getGradient());
    return dIn;
}

void Adjoint::write(std::ostream &out) const
{
    out << "sdpm::Adjoint on\n"
        << _sol;
}

#endif // USE_CODI
