/*
 * Copyright 2023 University of Liège
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#ifndef SDPMELEMENT_H
#define SDPMELEMENT_H

#include "sdpm.h"
#include "sdpmObject.h"
#include "sdpmNode.h"
#include "sdpmTag.h"
#include <vector>

namespace sdpm
{

/**
 * @brief Gmsh element type
 * @authors Adrien Crovato
 */
enum class ElType
{
    UNKNOWN = 0,
    LINE2 = 1,
    QUAD4 = 3
};
#ifndef SWIG
SDPM_API std::ostream &operator<<(std::ostream &out, ElType const &obj);
#endif

/**
 * @brief Element
 * @authors Adrien Crovato
 */
class SDPM_API Element : public sdpmObject
{
protected:
    // Mesh
    size_t _id;                 ///< element number
    Tag *_tag;                  ///< physical tag to which element belongs
    std::vector<Node *> _nodes; ///< nodes of the element
    bool _toDate;               ///< whether the element values are up-to-date
    // Geometry
    std::vector<sdpmVector3d> _coords; ///< scaled (compressible) coordinates
    sdpmVector3d _cg0;                 ///< element center of gravity
    sdpmDouble _srf0;                  ///< element surface area
    sdpmVector3d _nrm0;                ///< unit normal vector
    sdpmVector3d _cg;                  ///< compressible element center of gravity
    sdpmDouble _srf;                   ///< compressible element surface area
    sdpmVector3d _nrm;                 ///< compressible unit normal vector

public:
#ifndef SWIG
    Element(size_t n, Tag *tag, std::vector<Node *> &nodes);
    virtual ~Element() {}

#endif
    virtual ElType getType() const
    {
        return ElType::UNKNOWN;
    }

#ifndef SWIG
    // Management
    virtual void update();
    virtual void update(sdpmDouble beta);
    // Setters
    void setNode(size_t i, Node *n) { _nodes[i] = n; }
#endif
    // Getters
    size_t getId() const { return _id; }
    size_t const &getIdRef() const { return _id; }
    Tag const *getTag() const { return _tag; }
    std::vector<Node *> const &getNodes() const { return _nodes; }
    sdpmVector3d const &getCg() const { return _cg0; }
    sdpmDouble getSurfaceArea() const { return _srf0; }
    sdpmVector3d const &getNormal() const { return _nrm0; }
    inline std::vector<sdpmVector3d> const &getCompressibleCoords() const;
    inline sdpmVector3d const &getCompressibleCg() const;
    inline sdpmDouble getCompressibleSurfaceArea() const;
    inline sdpmVector3d const &getCompressibleNormal() const;
#ifndef SWIG
    // Utilities
    virtual sdpmMatrixXd computeGradientOperator() const;
    virtual sdpmVector3d computeGradient(std::vector<sdpmDouble> const &u) const;
    virtual void write(std::ostream &out) const override;
#endif
};

#ifndef SWIG
#include "sdpmElement.inl"
#endif

} // namespace sdpm

#endif // SDPMELEMENT_H
