/*
 * Copyright 2023 University of Liège
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include "sdpmGradient.h"
#include "sdpmSolver.h"
#include "sdpmProblem.h"
#include "sdpmBody.h"
#include <iostream>

using namespace sdpm;

Gradient::Gradient(Solver &sol) : _sol(sol),
                                  _nF(_sol._pbl._omega.size()),
                                  _nM(_sol._pbl.getNumModes())
{
    // Create node row identification map
    int inod = 0;
    for (auto body : _sol._pbl.getBodies())
        for (auto n : body->getNodes())
            if (_rowsN.insert(std::pair<Node *, int>(n, inod)).second == true)
                ++inod;
    _nN = _rowsN.size();

    // Resize displacement and load containers
    _x = sdpmVectorXd::Zero(_nN);
    _dz.resize(_nM, sdpmVectorXd::Zero(_nN));
    _clz1Re.resize(_nM, std::vector<sdpmVectorXd>(_nF, sdpmVectorXd::Zero(_nN)));
    _clz1Im.resize(_nM, std::vector<sdpmVectorXd>(_nF, sdpmVectorXd::Zero(_nN)));

    // Resize pressure gradient containers
    _dCp1Re_a.resize(_nF, sdpmVectorXd::Zero(_sol._nP));
    _dCp1Im_a.resize(_nF, sdpmVectorXd::Zero(_sol._nP));
    _dCp1Re_rx.resize(_nF, sdpmMatrixXd::Zero(_sol._nP, _sol._nP));
    _dCp1Im_rx.resize(_nF, sdpmMatrixXd::Zero(_sol._nP, _sol._nP));
    _dCp1Re_ry.resize(_nF, sdpmMatrixXd::Zero(_sol._nP, _sol._nP));
    _dCp1Im_ry.resize(_nF, sdpmMatrixXd::Zero(_sol._nP, _sol._nP));
    _dCp1Re_aDot.resize(_nF, sdpmVectorXd::Zero(_sol._nP));
    _dCp1Im_aDot.resize(_nF, sdpmVectorXd::Zero(_sol._nP));
    _dCp1Re_hDot.resize(_nF, sdpmVectorXd::Zero(_sol._nP));
    _dCp1Im_hDot.resize(_nF, sdpmVectorXd::Zero(_sol._nP));
    _dCp1Re_rxDot.resize(_nF, sdpmMatrixXd::Zero(_sol._nP, _sol._nP));
    _dCp1Im_rxDot.resize(_nF, sdpmMatrixXd::Zero(_sol._nP, _sol._nP));
    _dCp1Re_ryDot.resize(_nF, sdpmMatrixXd::Zero(_sol._nP, _sol._nP));
    _dCp1Im_ryDot.resize(_nF, sdpmMatrixXd::Zero(_sol._nP, _sol._nP));
    _dCp1Re_dzDot.resize(_nF, sdpmMatrixXd::Zero(_sol._nP, _sol._nP));
    _dCp1Im_dzDot.resize(_nF, sdpmMatrixXd::Zero(_sol._nP, _sol._nP));
    _dCp1Re_aDotDot.resize(_nF, sdpmVectorXd::Zero(_sol._nP));
    _dCp1Im_aDotDot.resize(_nF, sdpmVectorXd::Zero(_sol._nP));
    _dCp1Re_hDotDot.resize(_nF, sdpmVectorXd::Zero(_sol._nP));
    _dCp1Im_hDotDot.resize(_nF, sdpmVectorXd::Zero(_sol._nP));
    _dCp1Re_dzDotDot.resize(_nF, sdpmMatrixXd::Zero(_sol._nP, _sol._nP));
    _dCp1Im_dzDotDot.resize(_nF, sdpmMatrixXd::Zero(_sol._nP, _sol._nP));

    // Resize load gradient containers
    _dClz1Re_a.resize(_nF, sdpmVectorXd::Zero(_nN));
    _dClz1Im_a.resize(_nF, sdpmVectorXd::Zero(_nN));
    _dClz1Re_h.resize(_nF, sdpmVectorXd::Zero(_nN));
    _dClz1Im_h.resize(_nF, sdpmVectorXd::Zero(_nN));
    _dClz1Re_rx.resize(_nF, sdpmMatrixXd::Zero(_nN, _nN));
    _dClz1Im_rx.resize(_nF, sdpmMatrixXd::Zero(_nN, _nN));
    _dClz1Re_ry.resize(_nF, sdpmMatrixXd::Zero(_nN, _nN));
    _dClz1Im_ry.resize(_nF, sdpmMatrixXd::Zero(_nN, _nN));
    _dClz1Re_dz.resize(_nF, sdpmMatrixXd::Zero(_nN, _nN));
    _dClz1Im_dz.resize(_nF, sdpmMatrixXd::Zero(_nN, _nN));
}

Gradient::~Gradient()
{
    std::cout << "~sdpm::Gradient()\n";
}

/**
 * @brief Compute partial derivatives of pressure and loads with respect to degrees of freedom
 */
void Gradient::run()
{
    // Store coordinates, displacements and loads
    for (auto body : _sol._pbl.getBodies())
    {
        sdpmDouble xRef = _sol._pbl.getRefCtr()(0);
        std::vector<Node *> nods = body->getNodes();
        for (auto n : nods)
            _x(_rowsN.at(n)) = n->getCoords()(0) - xRef;
        for (size_t imd = 0; imd < _nM; ++imd)
        {
            sdpmDouble aAmp = body->getPitchAmplitude(imd);
            sdpmDouble hAmp = body->getPlungeAmplitude(imd);
            std::vector<sdpmDouble> dz = body->getModeZ(imd);
            for (auto n : nods)
                _dz[imd](_rowsN.at(n)) = dz[n->getId() - 1] + aAmp * _x(_rowsN.at(n)) + hAmp;
            for (size_t ifq = 0; ifq < _nF; ++ifq)
            {
                std::vector<sdpmVector3d> clzre = body->getUnsteadyLoadsReal(imd, ifq);
                std::vector<sdpmVector3d> clzim = body->getUnsteadyLoadsImag(imd, ifq);
                for (size_t i = 0; i < nods.size(); ++i)
                {
                    _clz1Re[imd][ifq](_rowsN.at(nods[i])) = clzre[i](2);
                    _clz1Im[imd][ifq](_rowsN.at(nods[i])) = clzim[i](2);
                }
            }
        }
    }

    // Surface gradient operator and average matrices
    std::vector<sdpmMatrixXd> dN;
    dN.resize(3, sdpmMatrixXd::Zero(_sol._rows.size(), _rowsN.size()));
    for (auto p : _sol._rows)
    {
        Element *e = p.first;
        std::vector<Node *> nodes = e->getNodes();
        sdpmMatrixXd dNe = e->computeGradientOperator();
        for (size_t in = 0; in < nodes.size(); ++in)
            for (size_t idim = 0; idim < 3; ++idim)
                dN[idim](_sol._rows.at(e), _rowsN.at(nodes[in])) = dNe(idim, in);
    }
    sdpmMatrixXd L = sdpmMatrixXd::Zero(_rowsN.size(), _sol._rows.size()); // elements-to-node average
    for (auto body : _sol._pbl.getBodies())
    {
        std::map<Node *, std::vector<Element *>> neMap = body->getMap();
        for (auto n : body->getNodes())
            for (auto e : neMap.at(n))
                L(_rowsN.at(n), _sol._rows.at(e)) = 1. / neMap.at(n).size();
    }
    sdpmMatrixXd L1 = sdpmMatrixXd::Zero(_sol._rows.size(), _rowsN.size()); // nodes-to-element average
    for (auto body : _sol._pbl.getBodies())
        for (auto e : body->getElements())
            for (auto n : e->getNodes())
                L1(_sol._rows.at(e), _rowsN.at(n)) = 1. / e->getNodes().size();

    // Freestream values
    sdpmDouble beta = _sol._pbl.getCompressibility();           // compressiblity factor
    sdpmDouble mach = _sol._pbl.getMach();                      // Mach number
    sdpmVector3d uinf = _sol._pbl.getDragDir();                 // velocity
    std::vector<sdpmDouble> omega = _sol._pbl.getFrequencies(); // circular frequencies
    // Geometry values
    sdpmVector3d xRef = _sol._pbl.getRefCtr(); // reference center of rotation
    // Organize panel and node data
    sdpmMatrixXd cg = sdpmMatrixXd::Zero(_sol._nP, 3); // panel CG
    sdpmMatrixXd s = sdpmVectorXd::Zero(_sol._nP);     // panel area
    sdpmMatrixXd n = sdpmMatrixXd::Zero(_sol._nP, 3);  // panel compressible normal
    sdpmMatrixXd nz = sdpmVectorXd::Zero(_sol._nP);    // panel z-normal
    sdpmMatrixXd u = sdpmMatrixXd::Zero(_sol._nP, 3);  // steady velocity
    sdpmVectorXd ux = sdpmVectorXd::Zero(_sol._nP);    // steady perturbation x-velocity
    for (auto body : _sol._pbl.getBodies())
    {
        for (auto e : body->getElements())
        {
            cg.row(_sol._rows.at(e)) = e->getCg();
            s(_sol._rows.at(e)) = e->getSurfaceArea();
            n.row(_sol._rows.at(e)) = e->getCompressibleNormal();
            nz(_sol._rows.at(e)) = e->getNormal()(2);
            u.row(_sol._rows.at(e)) = _sol._u[e->getId() - 1];
            ux(_sol._rows.at(e)) = _sol._u[e->getId() - 1](0) - uinf(0);
        }
    }

    // Compute derivatives
    std::cout << "Computing pressure derivatives... " << std::flush;
    for (size_t ifq = 0; ifq < omega.size(); ++ifq)
    {
        // K matrices
        sdpmMatrixXcd K = _sol._A[ifq].partialPivLu().inverse() * _sol._B[ifq];
        std::vector<sdpmMatrixXcd> Kx(3);
        for (size_t idim = 0; idim < 3; ++idim)
            Kx[idim] = dN[idim] * L * K;

        // Pressure stiffness terms (multiplied by 1)
        // w.r.t pitch
        sdpmVectorXcd cp_a = (2 / beta * u.col(0).array() - 2 * mach * mach / beta * ux.array()) * (uinf(0) * n.col(2).array() * n.col(0).array() + (Kx[0] * uinf(0) * n.col(2)).array()) + 2 * u.col(1).array() * (uinf(0) * n.col(2).array() * n.col(1).array() + (Kx[1] * uinf(0) * n.col(2)).array()) + 2 * u.col(2).array() * (-uinf(0) + uinf(0) * n.col(2).array() * n.col(2).array() + (Kx[2] * uinf(0) * n.col(2)).array());
        _dCp1Re_a[ifq] = cp_a.real();
        _dCp1Im_a[ifq] = cp_a.imag();
        // w.r.t modal x-rotations
        sdpmMatrixXcd cp_rx = (2 / beta * u.col(0).asDiagonal() - 2 * mach * mach / beta * ux.asDiagonal()) * (sdpmMatrixXd((uinf(2) * n.col(1) - uinf(1) * n.col(2)).asDiagonal()) * n.col(0).asDiagonal() + Kx[0] * (uinf(2) * n.col(1) - uinf(1) * n.col(2)).asDiagonal()) + 2 * u.col(1).asDiagonal() * (-uinf(2) * sdpmMatrixXd::Identity(_sol._nP, _sol._nP) + sdpmMatrixXd((uinf(2) * n.col(1) - uinf(1) * n.col(2)).asDiagonal()) * n.col(1).asDiagonal() + Kx[1] * (uinf(2) * n.col(1) - uinf(1) * n.col(2)).asDiagonal()) + 2 * u.col(2).asDiagonal() * (uinf(1) * sdpmMatrixXd::Identity(_sol._nP, _sol._nP) + sdpmMatrixXd((uinf(2) * n.col(1) - uinf(1) * n.col(2)).asDiagonal()) * n.col(2).asDiagonal() + Kx[2] * (uinf(2) * n.col(1) - uinf(1) * n.col(2)).asDiagonal());
        _dCp1Re_rx[ifq] = cp_rx.real();
        _dCp1Im_rx[ifq] = cp_rx.imag();
        // w.r.t modal y-rotations
        sdpmMatrixXcd cp_ry = -2 * u.col(0).asDiagonal() * uinf(2) * sdpmMatrixXd::Identity(_sol._nP, _sol._nP) + (2 / beta * u.col(0).asDiagonal() - 2 * mach * mach / beta * ux.asDiagonal()) * (sdpmMatrixXd((uinf(2) / beta * n.col(0) - uinf(0) * n.col(2)).asDiagonal()) * n.col(0).asDiagonal() + Kx[0] * (uinf(2) / beta * n.col(0) - uinf(0) * n.col(2)).asDiagonal()) + 2 * u.col(1).asDiagonal() * (sdpmMatrixXd((uinf(2) / beta * n.col(0) - uinf(0) * n.col(2)).asDiagonal()) * n.col(1).asDiagonal() + Kx[1] * (uinf(2) / beta * n.col(0) - uinf(0) * n.col(2)).asDiagonal()) + 2 * u.col(2).asDiagonal() * (uinf(0) * sdpmMatrixXd::Identity(_sol._nP, _sol._nP) + sdpmMatrixXd((uinf(2) / beta * n.col(0) - uinf(0) * n.col(2)).asDiagonal()) * n.col(2).asDiagonal() + Kx[2] * (uinf(2) / beta * n.col(0) - uinf(0) * n.col(2)).asDiagonal());
        _dCp1Re_ry[ifq] = cp_ry.real();
        _dCp1Im_ry[ifq] = cp_ry.imag();

        // Pressure damping terms (multiplied by i*omega)
        // w.r.t time derivative of pitch
        sdpmVectorXcd cp_aDot = 2 * u.col(0).array() * (cg.col(2).array() - xRef(2)) + (2 / beta * u.col(0).array() - 2 * mach * mach / beta * ux.array()) * ((-(cg.col(2).array() - xRef(2)) / beta * n.col(0).array() + (cg.col(0).array() - xRef(0)) * n.col(2).array()) * n.col(0).array() + (Kx[0] * (-(cg.col(2).array() - xRef(2)) / beta * n.col(0).array() + (cg.col(0).array() - xRef(0)) * n.col(2).array()).matrix()).array()) + 2 * u.col(1).array() * ((-(cg.col(2).array() - xRef(2)) / beta * n.col(0).array() + (cg.col(0).array() - xRef(0)) * n.col(2).array()) * n.col(1).array() + (Kx[1] * (-(cg.col(2).array() - xRef(2)) / beta * n.col(0).array() + (cg.col(0).array() - xRef(0)) * n.col(2).array()).matrix()).array()) + 2 * u.col(2).array() * (-(cg.col(0).array() - xRef(0)) + (-(cg.col(2).array() - xRef(2)) / beta * n.col(0).array() + (cg.col(0).array() - xRef(0)) * n.col(2).array()) * n.col(2).array() + (Kx[2] * (-(cg.col(2).array() - xRef(2)) / beta * n.col(0).array() + (cg.col(0).array() - xRef(0)) * n.col(2).array()).matrix()).array()) + 2 * (K * uinf(0) * n.col(2)).array() * (1 - mach * mach * ux.array());
        cp_aDot *= sdpmComplex(0., omega[ifq]);
        _dCp1Re_aDot[ifq] = cp_aDot.real();
        _dCp1Im_aDot[ifq] = cp_aDot.imag();
        // w.r.t time derivative of plunge
        sdpmVectorXcd cp_hDot = (2 / beta * u.col(0).array() - 2 * mach * mach / beta * ux.array()) * (n.col(2).array() * n.col(0).array() + (Kx[0] * n.col(2)).array()) + 2 * u.col(1).array() * (n.col(2).array() * n.col(1).array() + (Kx[1] * n.col(2)).array()) + 2 * u.col(2).array() * (-1 + n.col(2).array() * n.col(2).array() + (Kx[2] * n.col(2)).array());
        cp_hDot *= sdpmComplex(0., omega[ifq]);
        _dCp1Re_hDot[ifq] = cp_hDot.real();
        _dCp1Im_hDot[ifq] = cp_hDot.imag();
        // w.r.t time derivative of modal x-rotations
        sdpmMatrixXcd cp_rxDot = sdpmComplex(0., 2 * omega[ifq]) * (sdpmMatrixXd::Identity(_sol._nP, _sol._nP) - sdpmMatrixXd(mach * mach * ux.asDiagonal())) * (K * (uinf(2) * n.col(1) - uinf(1) * n.col(2)).asDiagonal());
        _dCp1Re_rxDot[ifq] = cp_rxDot.real();
        _dCp1Im_rxDot[ifq] = cp_rxDot.imag();
        // w.r.t time derivative of modal y-rotations
        sdpmMatrixXcd cp_ryDot = sdpmComplex(0., 2 * omega[ifq]) * (sdpmMatrixXd::Identity(_sol._nP, _sol._nP) - sdpmMatrixXd(mach * mach * ux.asDiagonal())) * (K * (uinf(2) / beta * n.col(0) - uinf(0) * n.col(2)).asDiagonal());
        _dCp1Re_ryDot[ifq] = cp_ryDot.real();
        _dCp1Im_ryDot[ifq] = cp_ryDot.imag();
        // w.r.t time derivative of modal z-displacements
        sdpmMatrixXcd cp_dzDot = (2 / beta * u.col(0).asDiagonal() - 2 * mach * mach / beta * ux.asDiagonal()) * (sdpmMatrixXd(n.col(2).asDiagonal()) * n.col(0).asDiagonal() + Kx[0] * n.col(2).asDiagonal()) + 2 * u.col(1).asDiagonal() * (sdpmMatrixXd(n.col(2).asDiagonal()) * n.col(1).asDiagonal() + Kx[1] * n.col(2).asDiagonal()) + 2 * u.col(2).asDiagonal() * (-sdpmMatrixXd::Identity(_sol._nP, _sol._nP) + sdpmMatrixXd(n.col(2).asDiagonal()) * n.col(2).asDiagonal() + Kx[2] * n.col(2).asDiagonal());
        cp_dzDot *= sdpmComplex(0., omega[ifq]);
        _dCp1Re_dzDot[ifq] = cp_dzDot.real();
        _dCp1Im_dzDot[ifq] = cp_dzDot.imag();

        // Pressure inertia terms (multiplied by (i*omega)^2)
        // w.r.t second time derivative of pitch
        sdpmVectorXcd cp_aDotDot = -2 * omega[ifq] * omega[ifq] * (1 - mach * mach * ux.array()) * (K * (-(cg.col(2).array() - xRef(2)) / beta * n.col(0).array() + (cg.col(0).array() - xRef(0)) * n.col(2).array()).matrix()).array();
        _dCp1Re_aDotDot[ifq] = cp_aDotDot.real();
        _dCp1Im_aDotDot[ifq] = cp_aDotDot.imag();
        // w.r.t second time derivative of plunge
        sdpmVectorXcd cp_hDotDot = -2 * omega[ifq] * omega[ifq] * (1 - mach * mach * ux.array()) * (K * n.col(2)).array();
        _dCp1Re_hDotDot[ifq] = cp_hDotDot.real();
        _dCp1Im_hDotDot[ifq] = cp_hDotDot.imag();
        // w.r.t second time derivative of modal z-displacements
        sdpmMatrixXcd cp_dzDotDot = -2 * omega[ifq] * omega[ifq] * (sdpmMatrixXd::Identity(_sol._nP, _sol._nP) - sdpmMatrixXd(mach * mach * ux.asDiagonal())) * (K * n.col(2).asDiagonal());
        _dCp1Re_dzDotDot[ifq] = cp_dzDotDot.real();
        _dCp1Im_dzDotDot[ifq] = cp_dzDotDot.imag();

        // Load terms
        sdpmVectorXcd clz_a = L1.transpose() * s.asDiagonal() * nz.asDiagonal() * (cp_a + cp_aDot + cp_aDotDot);
        _dClz1Re_a[ifq] = clz_a.real();
        _dClz1Im_a[ifq] = clz_a.imag();
        sdpmVectorXcd clz_h = L1.transpose() * s.asDiagonal() * nz.asDiagonal() * (cp_hDot + cp_hDotDot);
        _dClz1Re_h[ifq] = clz_h.real();
        _dClz1Im_h[ifq] = clz_h.imag();
        sdpmMatrixXcd clz_rx = L1.transpose() * s.asDiagonal() * nz.asDiagonal() * (cp_rx + cp_rxDot) * L1;
        _dClz1Re_rx[ifq] = clz_rx.real();
        _dClz1Im_rx[ifq] = clz_rx.imag();
        sdpmMatrixXcd clz_ry = L1.transpose() * s.asDiagonal() * nz.asDiagonal() * (cp_ry + cp_ryDot) * L1;
        _dClz1Re_ry[ifq] = clz_ry.real();
        _dClz1Im_ry[ifq] = clz_ry.imag();
        sdpmMatrixXcd clz_dz = L1.transpose() * s.asDiagonal() * nz.asDiagonal() * (cp_dzDot + cp_dzDotDot) * L1;
        _dClz1Re_dz[ifq] = clz_dz.real();
        _dClz1Im_dz[ifq] = clz_dz.imag();
    }
    std::cout << "done." << std::endl;
}

/**
 * @brief Compute partial derivative of real part of GAF matrix entry at given frequency with respect to given mode
 */
std::vector<sdpmDouble> Gradient::computePartialsGafRe(size_t ifq, int imd, int irw, int jcl, std::string const &wrt)
{
    // Check input
    sdpmMatrixXd dClz1;
    if (wrt == "a")
        dClz1 = _dClz1Re_a[ifq];
    else if (wrt == "h")
        dClz1 = _dClz1Re_h[ifq];
    else if (wrt == "rx")
        dClz1 = _dClz1Re_rx[ifq];
    else if (wrt == "ry")
        dClz1 = _dClz1Re_ry[ifq];
    else if (wrt == "dz")
        dClz1 = _dClz1Re_dz[ifq];
    else
        throw std::runtime_error("sdpm::Gradient::computePartialsGafRe: wrt parameter should be either 'a', 'h', 'rx', 'ry' or 'dz'!\n");

    // Compute derivative
    sdpmVectorXd dq;
    if (wrt == "a" || wrt == "h")
        dq = sdpmVectorXd::Zero(1);
    else
        dq = sdpmVectorXd::Zero(_nN);
    if (jcl == imd)
        dq += dClz1.transpose() * _dz[irw];
    if (irw == imd)
    {
        if (wrt == "a")
            dq -= _x.transpose() * _clz1Re[jcl][ifq];
        else if (wrt == "h")
            dq -= sdpmVectorXd::Constant(_nN, 1.0).transpose() * _clz1Re[jcl][ifq];
        else if (wrt == "dz")
            dq -= _clz1Re[jcl][ifq];
    }
    return std::vector<sdpmDouble>(dq.data(), dq.data() + dq.size());
}

/**
 * @brief Compute partial derivative of imaginary part of GAF matrix entry at given frequency with respect to given mode
 */
std::vector<sdpmDouble> Gradient::computePartialsGafIm(size_t ifq, int imd, int irw, int jcl, std::string const &wrt)
{
    // Check input
    sdpmMatrixXd dClz1;
    if (wrt == "a")
        dClz1 = _dClz1Im_a[ifq];
    else if (wrt == "h")
        dClz1 = _dClz1Im_h[ifq];
    else if (wrt == "rx")
        dClz1 = _dClz1Im_rx[ifq];
    else if (wrt == "ry")
        dClz1 = _dClz1Im_ry[ifq];
    else if (wrt == "dz")
        dClz1 = _dClz1Im_dz[ifq];
    else
        throw std::runtime_error("sdpm::Gradient::computePartialsGafIm: wrt parameter should be either 'a', 'h', 'rx', 'ry' or 'dz'!\n");

    // Compute derivative
    sdpmVectorXd dq;
    if (wrt == "a" || wrt == "h")
        dq = sdpmVectorXd::Zero(1);
    else
        dq = sdpmVectorXd::Zero(_nN);
    if (jcl == imd)
        dq += dClz1.transpose() * _dz[irw];
    if (irw == imd)
    {
        if (wrt == "a")
            dq -= _x.transpose() * _clz1Im[jcl][ifq];
        else if (wrt == "h")
            dq -= sdpmVectorXd::Constant(_nN, 1.0).transpose() * _clz1Im[jcl][ifq];
        else if (wrt == "dz")
            dq -= _clz1Im[jcl][ifq];
    }
    return std::vector<sdpmDouble>(dq.data(), dq.data() + dq.size());
}

void Gradient::write(std::ostream &out) const
{
    out << "sdpm::Gradient on\n"
        << _sol;
}
