/*
 * Copyright 2023 University of Liège
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#ifndef SDPMPROBLEM_H
#define SDPMPROBLEM_H

#include "sdpm.h"
#include "sdpmObject.h"
#include "sdpmMesh.h"
#include <vector>

namespace sdpm
{

/**
 * @brief Hold the problem definition
 * @authors Adrien Crovato
 */
class SDPM_API Problem : public sdpmObject
{
    friend class Adjoint;  // so that Adjoint can access the problem inputs
    friend class Gradient; // so that Gradient can access the problem outputs

    // Mesh
    Mesh &_msh;                  ///< mesh data structure
    std::vector<Body *> _bodies; ///< bodies in fluid
    // Aerodynamic
    sdpmDouble _aoa;    ///< Angle of attack
    sdpmDouble _aos;    ///< Angle of sideslip
    sdpmDouble _mach;   ///< Mach number
    sdpmDouble _beta;   ///< Compressibility factor
    sdpmVector3d _dDir; ///< Direction of drag force
    sdpmVector3d _sDir; ///< Direction of side force
    sdpmVector3d _lDir; ///< Direction of lift force
    // Geometry
    sdpmDouble _sRef;   ///< Reference surface area
    sdpmDouble _cRef;   ///< Reference chord length
    sdpmVector3d _xRef; ///< Reference center point for moment computation
    bool _ySym;         ///< True if mesh is symmetric with respect to y-plane
    // Unsteady
    size_t _nModes;                 ///< Number of modal motions
    std::vector<sdpmDouble> _omega; ///< Circular frequencies
    // Check flags
    bool _hasAero; ///< True if aerodynamic parameters have been updated
    bool _hasGeo;  ///< True if geometric parameters have been updated

public:
    Problem(Mesh &msh);
    ~Problem();

#ifndef SWIG
    std::vector<Body *> const &getBodies() const
    {
        return _bodies;
    }
    std::vector<Node *> const &getNodes() const { return _msh.getNodes(); }
    std::vector<Element *> const &getElements() const { return _msh.getElements(); }
    sdpmDouble getAoa() const { return _aoa; }
    sdpmDouble getAos() const { return _aos; }
    sdpmDouble getMach() const { return _mach; }
    sdpmDouble getCompressibility() const { return _beta; }
    sdpmVector3d const &getDragDir() const { return _dDir; }
    sdpmVector3d const &getSideDir() const { return _sDir; }
    sdpmVector3d const &getLiftDir() const { return _lDir; }
    sdpmDouble getRefArea() const { return _sRef; }
    sdpmDouble getRefLgth() const { return _cRef; }
    sdpmVector3d const &getRefCtr() const { return _xRef; }
    bool getSym() const { return _ySym; }
    size_t getNumModes() const { return _nModes; }
    std::vector<sdpmDouble> const &getFrequencies() const { return _omega; }
#endif
    void addBody(Body &b);
    void setFreestream(double aoa, double aos, double mch);
    void setGeometry(double sref, double cref, double xref, double yref, double zref, bool ysym);
    void setUnsteady(size_t nmods, std::vector<double> const &ks);
#ifndef SWIG
    void check() const;
    void update();
    virtual void write(std::ostream &out) const override;
#endif
};

} // namespace sdpm

#endif // SDPMPROBLEM_H
