/*
 * Copyright 2023 University of Liège
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#ifndef SDPMSOLVER_H
#define SDPMSOLVER_H

#include "sdpm.h"
#include "sdpmObject.h"
#include <vector>
#include <map>

namespace sdpm
{

/**
 * @brief Source-doublet panel solver
 * @authors Adrien Crovato
 */
class SDPM_API Solver : public sdpmObject
{
    friend class Adjoint;  // so that Adjoint can access the solver outputs
    friend class Gradient; // so that Gradient can access the solver outputs

protected:
    // Problem
    Problem &_pbl; ///< problem definition
    int _vrb;      ///< verbosity level
    // AICs
    size_t _nP;                     ///< number of body panels
    std::map<Element *, int> _rows; ///< map linking an element to a matrix entry
    sdpmMatrixXd _A0;               ///< zero-frequency body-body doublet matrix
    sdpmMatrixXd _B0;               ///< zero-frequency body-body source matrix
    std::vector<sdpmMatrixXcd> _A;  ///< unsteady body-body doublet matrix
    std::vector<sdpmMatrixXcd> _B;  ///< unsteady body-body source matrix

private:
    // Solution
    std::vector<sdpmDouble> _tau;                              ///< source strength
    std::vector<sdpmDouble> _mu;                               ///< doublet strength (perturbation potential)
    std::vector<sdpmVector3d> _u;                              ///< steady velocity
    std::vector<sdpmDouble> _cp;                               ///< steady pressure coefficient
    std::vector<std::vector<std::vector<sdpmVector3d>>> _u1Re; ///< real part of unsteady first harmonic velocity
    std::vector<std::vector<std::vector<sdpmVector3d>>> _u1Im; ///< imaginary part of unsteady first harmonic velocity
    std::vector<std::vector<std::vector<sdpmDouble>>> _cp1Re;  ///< real part of unsteady first harmonic pressure coefficient
    std::vector<std::vector<std::vector<sdpmDouble>>> _cp1Im;  ///< imaginary part of unsteady first harmonic pressure coefficient
    // Aerodynamic loads
    sdpmDouble _cl;                              ///< steady lift coefficient
    sdpmDouble _cd;                              ///< steady drag coefficient
    sdpmDouble _cs;                              ///< steady sideforce coefficient
    sdpmDouble _cm;                              ///< steady pitch moment coefficient (positive nose-up)
    std::vector<std::vector<sdpmDouble>> _cl1Re; ///< real part of unsteady first harmonic lift coefficient
    std::vector<std::vector<sdpmDouble>> _cl1Im; ///< imaginary part of unsteady first harmonic lift coefficient
    std::vector<std::vector<sdpmDouble>> _cd1Re; ///< real part of unsteady first harmonic drag coefficient
    std::vector<std::vector<sdpmDouble>> _cd1Im; ///< imaginary part of unsteady first harmonic drag coefficient
    std::vector<std::vector<sdpmDouble>> _cs1Re; ///< real part of unsteady first harmonic side force coefficient
    std::vector<std::vector<sdpmDouble>> _cs1Im; ///< imaginary part of unsteady first harmonic side force coefficient
    std::vector<std::vector<sdpmDouble>> _cm1Re; ///< real part of unsteady first harmonic pitch moment coefficient (positive nose-up)
    std::vector<std::vector<sdpmDouble>> _cm1Im; ///< imaginary part of unsteady first harmonic pitch moment coefficient (positive nose-up)
    // Generalized Aerodynamic Force matrices
    std::vector<sdpmMatrixXd> _q1Re; ///< real part of GAF matrices
    std::vector<sdpmMatrixXd> _q1Im; ///< imaginary part of GAF matrices

public:
    Solver(Problem &pbl, int vrb = 1);
    virtual ~Solver();

    std::vector<sdpmDouble> const &getPressure() const { return _cp; }
    std::vector<sdpmDouble> const &getUnsteadyPressureReal(size_t imd, size_t ifq) const { return _cp1Re[imd][ifq]; }
    std::vector<sdpmDouble> const &getUnsteadyPressureImag(size_t imd, size_t ifq) const { return _cp1Im[imd][ifq]; }
    sdpmDouble getLiftCoeff() const { return _cl; }
    sdpmDouble getDragCoeff() const { return _cd; }
    sdpmDouble getSideforceCoeff() const { return _cs; }
    sdpmDouble getMomentCoeff() const { return _cm; }
    sdpmDouble getUnsteadyLiftCoeffReal(size_t imd, size_t ifq) const { return _cl1Re[imd][ifq]; }
    sdpmDouble getUnsteadyLiftCoeffImag(size_t imd, size_t ifq) const { return _cl1Im[imd][ifq]; }
    sdpmDouble getUnsteadyDragCoeffReal(size_t imd, size_t ifq) const { return _cd1Re[imd][ifq]; }
    sdpmDouble getUnsteadyDragCoeffImag(size_t imd, size_t ifq) const { return _cd1Im[imd][ifq]; }
    sdpmDouble getUnsteadySideforceCoeffReal(size_t imd, size_t ifq) const { return _cs1Re[imd][ifq]; }
    sdpmDouble getUnsteadySideforceCoeffImag(size_t imd, size_t ifq) const { return _cs1Im[imd][ifq]; }
    sdpmDouble getUnsteadyMomentCoeffReal(size_t imd, size_t ifq) const { return _cm1Re[imd][ifq]; }
    sdpmDouble getUnsteadyMomentCoeffImag(size_t imd, size_t ifq) const { return _cm1Im[imd][ifq]; }
    std::vector<std::vector<sdpmDouble>> getGafMatrixRe(size_t ifq) const;
    std::vector<std::vector<sdpmDouble>> getGafMatrixIm(size_t ifq) const;

    void update();
    virtual void run();
    void save(GmshExport const &writer, std::string const &suffix = "");

#ifndef SWIG
    virtual void write(std::ostream &out) const override;
#endif

protected:
    void post(sdpmVectorXd const &etau, sdpmVectorXd const &emu);
    void post(size_t imd, size_t ifq, sdpmVectorXcd const &etau, sdpmVectorXcd const &emu);
    void computeGAF(size_t ifq);
};

} // namespace sdpm
#endif // SDPMSOLVER_H
