/*
 * Copyright 2023 University of Liège
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include "sdpmSolverTransonic.h"
#include "sdpmProblem.h"
#include "sdpmBody.h"
#include "sdpm_ad_interface.h"
#include <iostream>

using namespace sdpm;

SolverTransonic::SolverTransonic(Problem &pbl, int vrb) : Solver(pbl, vrb), _dCorr(sdpmVectorXd::Ones(_nP))
{
    // Build node row id
    int i = 0;
    for (auto body : _pbl.getBodies())
        for (auto n : body->getNodes())
            if (_rowsN.insert(std::pair<Node *, int>(n, i)).second == true)
                ++i;
}

/**
 * @brief Solve the steady and unsteady potential equations corrected for transonic effects
 */
void SolverTransonic::run()
{
    std::cout << "Computing transonic correction...";
    computeCorrection();
    std::cout << "done." << std::endl;

    std::cout << "Running transonic SDPM solver..." << std::flush;
    if (_vrb > 0)
        std::cout << "\n-- Mean flow\n"
                  << "computing sources... " << std::flush;
    sdpmDouble beta = _pbl.getCompressibility();
    sdpmVector3d ui = _pbl.getDragDir();
    sdpmVectorXd etauxy(_nP);
    sdpmVectorXd etauz(_nP);
    for (auto body : _pbl.getBodies())
    {
        for (auto e : body->getElements())
        {
            etauxy(_rows[e]) = ui[0] * e->getCompressibleNormal()[0] / beta + ui[1] * e->getCompressibleNormal()[1];
            etauz(_rows[e]) = ui[2] * e->getCompressibleNormal()[2];
        }
    }
    if (_vrb > 0)
        std::cout << "done.\n"
                  << "solving for doublets... " << std::flush;
    sdpmVectorXd rhsxy = _B0 * etauxy;
    sdpmVectorXd rhsz = _B0 * etauz;
    sdpmVectorXd emuxy = solve(_A0, rhsxy);
    sdpmVectorXd emuz = solve(_A0, rhsz);
    if (_vrb > 0)
        std::cout << "done.\n"
                  << "computing flow and loads on bodies... " << std::flush;
    post(etauxy + etauz, emuxy + emuz.cwiseProduct(_dCorr));
    if (_vrb > 0)
        std::cout << "done." << std::endl;

    for (size_t imd = 0; imd < _pbl.getNumModes(); ++imd)
    {
        for (size_t ifq = 0; ifq < _pbl.getFrequencies().size(); ++ifq)
        {
            if (_vrb > 0)
                std::cout << "-- Mode #" << imd << ", frequency #" << ifq << "\n"
                          << "computing sources... " << std::flush;
            sdpmVectorXcd etau1(_nP);
            sdpmVectorXcd etau1xy(_nP);
            sdpmVectorXcd etau1z(_nP);
            for (auto body : _pbl.getBodies())
            {
                std::vector<sdpmVector3d> uc0 = body->getSteadyVelocity(imd);
                std::vector<sdpmVector3d> uc1 = body->getHarmonicVelocity(imd);
                std::vector<Element *> elems = body->getElements();
                for (size_t i = 0; i < elems.size(); ++i)
                    etau1(_rows[elems[i]]) = (uc0[i] + uc1[i] * sdpmComplex(0., 1.) * _pbl.getFrequencies()[ifq]).conjugate().dot(elems[i]->getCompressibleNormal().cwiseQuotient(sdpmVector3d(_pbl.getCompressibility(), 1., 1.)));
                for (size_t i = 0; i < elems.size(); ++i)
                {
                    etau1xy(_rows[elems[i]]) = (uc0[i](0) + uc1[i](0) * sdpmComplex(0., 1.) * _pbl.getFrequencies()[ifq]) * elems[i]->getCompressibleNormal()(0) / beta + (uc0[i](1) + uc1[i](1) * sdpmComplex(0., 1.) * _pbl.getFrequencies()[ifq]) * elems[i]->getCompressibleNormal()(1);
                    etau1z(_rows[elems[i]]) = (uc0[i](2) + uc1[i](2) * sdpmComplex(0., 1.) * _pbl.getFrequencies()[ifq]) * elems[i]->getCompressibleNormal()(2);
                }
            }
            if (_vrb > 0)
                std::cout << "done.\n"
                          << "solving for doublets... " << std::flush;
            // Split complex set of equations into real set (twice the size)
            sdpmMatrixXd A(2 * _nP, 2 * _nP);
            A.block(0, 0, _nP, _nP) = _A[ifq].real();
            A.block(_nP, _nP, _nP, _nP) = _A[ifq].real();
            A.block(_nP, 0, _nP, _nP) = _A[ifq].imag();
            A.block(0, _nP, _nP, _nP) = -_A[ifq].imag();
            sdpmVectorXd rhs1xy(2 * _nP);
            sdpmVectorXd rhs1z(2 * _nP);
            rhs1xy.block(0, 0, _nP, 1) = (_B[ifq] * etau1xy).real();
            rhs1xy.block(_nP, 0, _nP, 1) = (_B[ifq] * etau1xy).imag();
            rhs1z.block(0, 0, _nP, 1) = (_B[ifq] * etau1z).real();
            rhs1z.block(_nP, 0, _nP, 1) = (_B[ifq] * etau1z).imag();
            sdpmVectorXd muxy = solve(A, rhs1xy);
            sdpmVectorXd muz = solve(A, rhs1z);
            sdpmVectorXcd emu1xy = muxy.block(0, 0, _nP, 1) + sdpmComplex(0, 1) * muxy.block(_nP, 0, _nP, 1);
            sdpmVectorXcd emu1z = muz.block(0, 0, _nP, 1) + sdpmComplex(0, 1) * muz.block(_nP, 0, _nP, 1);
            if (_vrb > 0)
                std::cout << "done.\n"
                          << "computing flow and loads on bodies... " << std::flush;
            sdpmVectorXcd mu1c = emu1xy + _dCorr.cwiseProduct(emu1z);
            post(imd, ifq, etau1, mu1c);
            if (_vrb > 0)
                std::cout << "done." << std::endl;
        }
    }

    // GAF
    if (_vrb > 0)
        std::cout << "-- Generalized Aerodynamic Force matrices" << std::endl;
    for (size_t ifq = 0; ifq < _pbl.getFrequencies().size(); ++ifq)
    {
        if (_vrb > 0)
            std::cout << "computing GAF at frequency #" << ifq << "... " << std::flush;
        computeGAF(ifq);
        if (_vrb > 0)
            std::cout << "done." << std::endl;
    }
    if (_vrb == 0)
        std::cout << " done.";
    std::cout << std::endl;
}

/**
 * @brief Compute the transonic correction terms
 */
void SolverTransonic::computeCorrection()
{
    // Assemble x-gradient matrix
    // 1) Gradient matrix on element
    sdpmMatrixXd dNx = sdpmMatrixXd::Zero(_rows.size(), _rowsN.size());
    for (auto p : _rows)
    {
        Element *e = p.first;
        std::vector<Node *> nodes = e->getNodes();
        sdpmVectorXd dNe = e->computeGradientOperator().row(0).transpose();
        for (size_t in = 0; in < nodes.size(); ++in)
            dNx(_rows.at(e), _rowsN.at(nodes[in])) = dNe(in);
    }
    // 2) Average matrix from element to node
    sdpmMatrixXd L = sdpmMatrixXd::Zero(_rowsN.size(), _rows.size());
    for (auto body : _pbl.getBodies())
    {
        std::map<Node *, std::vector<Element *>> neMap = body->getMap();
        for (auto n : body->getNodes())
            for (auto e : neMap.at(n))
                L(_rowsN.at(n), _rows.at(e)) = 1. / neMap.at(n).size();
    }
    sdpmMatrixXd dNxL = dNx * L;
    // 3) Boundary condition on upper trailing edge
    for (auto body : _pbl.getBodies())
    {
        std::vector<Element *> teElems;
        std::tie(teElems, std::ignore) = body->getTrailingEdgeElements();
        for (auto e : teElems)
        {
            dNxL.row(_rows.at(e)) = sdpmVectorXd::Zero(_nP).transpose();
            dNxL(_rows.at(e), _rows.at(e)) = 1.0;
        }
    }

    // Assemble RHS
    // 1) Average derivative of pressure wrt angle of attack at element
    sdpmVectorXd dCpElems = sdpmVectorXd::Zero(_nP);
    for (auto body : _pbl.getBodies())
    {
        std::vector<sdpmDouble> dCpNodes = body->getTransonicPressureGrad();
        for (auto e : body->getElements())
            for (auto n : e->getNodes())
                dCpElems(_rows.at(e)) += dCpNodes[n->getId() - 1] / e->getNodes().size();
    }
    // 2) Compute derivative of doublet wrt angle of attack
    sdpmVectorXd nz(_nP);
    for (auto body : _pbl.getBodies())
        for (auto e : body->getElements())
            nz(_rows[e]) = e->getCompressibleNormal()[2] * cos(_pbl.getAos());
    sdpmVectorXd rhs = _B0 * nz;
    sdpmVectorXd dMuA = solve(_A0, rhs);
    // 3) Compute RHS with boundary condition
    sdpmVectorXd dMuAxCorr = sdpmVectorXd::Zero(_nP);
    for (auto body : _pbl.getBodies())
    {
        for (auto e : body->getElements())
            dMuAxCorr(_rows.at(e)) = 0.5 * _pbl.getCompressibility() * dCpElems(_rows.at(e)) - nz(_rows[e]) * e->getCompressibleNormal()[0];
        std::vector<Element *> teElems;
        std::tie(teElems, std::ignore) = body->getTrailingEdgeElements();
        for (auto e : teElems)
            dMuAxCorr(_rows.at(e)) = dMuA(_rows.at(e));
    }

    // Solve for correction
    sdpmVectorXd dMuACorr = solve(dNxL, dMuAxCorr);
    _dCorr = dMuACorr.cwiseQuotient(dMuA);
}

void SolverTransonic::write(std::ostream &out) const
{
    out << "sdpm::SolverTransonic with\n"
        << _pbl;
}
