/*
 * Copyright 2023 University of Liège
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#ifndef SDPMTIMERS_H
#define SDPMTIMERS_H

#include "sdpm.h"
#include "sdpmObject.h"
#include <map>
#include <chrono>
#include <ctime>
#include <string>

namespace sdpm
{

/**
 * @brief Basic and portable timer
 * @author Adrien Crovato
 */
class SDPM_API Timer
{
private:
    bool _isRunning;                                    ///< status
    std::chrono::high_resolution_clock::time_point _tS; ///< start time (wall-clock)
    std::chrono::high_resolution_clock::time_point _tE; ///< stop time (wall-clock)
    std::clock_t _cS;                                   ///< start time (cpu)
    std::clock_t _cE;                                   ///< stop time (cpu)

public:
    Timer();
    ~Timer() {}

    void tick(std::chrono::high_resolution_clock::time_point &tp, std::clock_t &ct);
    void start();
    void stop();
    double getWall() const { return std::chrono::duration<double>(_tE - _tS).count(); }
    double getCpu() const { return (_cE - _cS) / (double)CLOCKS_PER_SEC; }
};

/**
 * @brief Collection of timers
 * @author Adrien Crovato
 */
class SDPM_API Timers : public sdpmObject
{
private:
    std::map<std::string, Timer> _tms; ///< set of timers

public:
    Timers() : sdpm::sdpmObject() {}
    ~Timers() {}

#ifndef SWIG
    Timer &operator[](std::string const &name);
    virtual void write(std::ostream &out) const override;
#endif // SWIG
};

} // namespace sdpm

#endif // SDPMTIMERS_H
