/*
 * Copyright 2023 University of Liège
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include "sdpmWake.h"
#include "sdpmEdge.h"
#include "sdpmMesh.h"
#include "sdpmNode.h"
#include "sdpmElement.h"
#include <vector>
#include <unordered_set>
using namespace sdpm;

Wake::Wake(Mesh &msh, std::string const &name, Tag *const &wingTag, size_t nTe, std::map<Node *, Node *> const &iTeMap) : Group(msh, name)
{
    // Create TE edges
    std::vector<Element *> elems = _tag->getElements();
    std::unordered_set<Edge *, EquEdge, EquEdge> teEdges;
    for (size_t i = 0; i < nTe; ++i)
    {
        std::vector<Node *> nods = elems[i]->getNodes();
        Edge *ed = new Edge({nods[0], nods[3]});
        ed->el0 = elems[i];
        teEdges.insert(ed);
    }

    // Save nodes
    std::vector<Node *> mshNodes = msh.getNodes();
    size_t endNodW = mshNodes.size();
    size_t nNodWake = ((elems.size() / nTe) + 1) * (nTe + 1);
    _nodes.resize(nNodWake);
    // first nTe+1 nodes are trailing edge nodes
    for (size_t i = 0; i <= nTe; ++i)
        _nodes[i] = elems[i]->getNodes()[0];
    _nodes[nTe] = elems[nTe - 1]->getNodes()[3];
    // other nodes are last added nodes in the mesh (other than duplicated trailing edge nodes)
    for (size_t i = 1; i <= nNodWake - nTe - 1; ++i)
        _nodes[nNodWake - i] = mshNodes[endNodW - i];

    // Find wing elements having an edge on the TE...
    // ... the elements on the pressure side still have their original (suction side) trailing edge nodes
    if (iTeMap.empty())
    {
        for (auto e : wingTag->getElements())
        {
            std::vector<Node *> nods = e->getNodes();
            size_t idx = 0;
            for (size_t i = 0; i < nods.size(); ++i)
            {
                // build wing edge and check if wing edge is a TE edge
                std::vector<Node *> eN(2);
                for (size_t j = 0; j < 2; ++j)
                    eN[j] = nods[(idx + j) % nods.size()];
                Edge ed(eN);
                auto it = teEdges.find(&ed);
                if (it != teEdges.end())
                {
                    // check normal direction
                    if (e->getNormal().dot((*it)->el0->getNormal()) > 0)
                        (*it)->el1 = e; // suction side
                    else
                        (*it)->el2 = e; // pressure side
                    break;
                }
                idx++;
            }
        }
    }
    // ... the elements on the pressure side already have their correct (pressure side) trailing edge nodes
    else
    {
        for (auto e : wingTag->getElements())
        {
            std::vector<Node *> nods = e->getNodes();
            size_t idx = 0;
            for (size_t i = 0; i < nods.size(); ++i, ++idx)
            {
                // build wing edge and check if wing edge is a TE edge
                std::vector<Node *> eN(2);
                for (size_t j = 0; j < 2; ++j)
                    eN[j] = nods[(idx + j) % nods.size()];
                Edge ed(eN);
                auto it = teEdges.find(&ed);
                // try with actual nodes
                if (it != teEdges.end())
                {
                    (*it)->el1 = e; // suction side
                    break;
                }
                // otherwise try again, but by replacing the nodes by their corresponding upper nodes
                else
                {
                    // build wing edge and check if wing edge is a TE edge
                    std::vector<Node *> eN(2);
                    try
                    {
                        for (size_t j = 0; j < 2; ++j)
                            eN[j] = iTeMap.at(nods[(idx + j) % nods.size()]);
                    }
                    catch (const std::out_of_range &)
                    {
                        continue;
                    }
                    Edge ed(eN);
                    auto it = teEdges.find(&ed);
                    if (it != teEdges.end())
                    {
                        (*it)->el2 = e; // pressure side
                        break;
                    }
                }
            }
        }
    }

    // Map wake elements on the TE to wing elements
    for (auto ed : teEdges)
    {
        if (ed->el1 == NULL || ed->el2 == NULL)
        {
            std::stringstream err;
            err << "sdpm::Wake: at least one of the two trailing edge wing elements from tag " << *wingTag << "could not be associated to wake element " << *(ed->el0) << "!\n";
            throw std::runtime_error(err.str());
        }
        _wwMap[ed->el0] = std::pair<Element *, Element *>(ed->el1, ed->el2);
        delete ed;
    }
    // Map the rest of the wake and fill time lag map
    sdpmDouble dt = elems[0]->getNodes()[1]->getCoords()(0) - elems[0]->getNodes()[0]->getCoords()(0); // dt = dx / ux, (ux~1)
    size_t nw = elems.size() / nTe;                                                                    // number of shed wake rows
    for (size_t j = 0; j < nTe; ++j)
    {
        _lagMap[elems[j]] = dt;
        std::pair<Element *, Element *> p = _wwMap.at(elems[j]);
        for (size_t i = 1; i < nw; ++i)
        {
            _wwMap[elems[i * nTe + j]] = p;
            _lagMap[elems[i * nTe + j]] = (i + 1) * dt;
        }
    }
}

void Wake::write(std::ostream &out) const
{
    out << "sdpm::Wake " << *_tag;
}
