/*
 * Copyright 2023 University of Liège
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#ifndef SDPM_AD_INTERFACE_H
#define SDPM_AD_INTERFACE_H

#include "sdpm.h"

namespace sdpm
{
#ifndef USE_CODI
// Getter
inline double getValue(const sdpmDouble &d) { return d; }

// Dummy pre-accumulation helper
inline void startPreAcc() {}
inline void stopPreAcc() {}
inline void addPreAccIn(sdpmDouble const *v, size_t size = 1) {}
inline void addPreAccOut(sdpmDouble *v, size_t size = 1) {}

// Solver helper function
inline sdpmVectorXd solve(sdpmMatrixXd &A, sdpmVectorXd &b)
{
    return A.partialPivLu().solve(b);
}

#else // USE_CODI
// Getter
inline double getValue(const sdpmDouble &d) { return d.getValue(); }

// Pre-accumulation helper
extern codi::PreaccumulationHelper<sdpmDouble> preAccHlpr;
inline void startPreAcc()
{
    if (sdpmDouble::getTape().isActive())
        preAccHlpr.start();
}
inline void stopPreAcc()
{
    if (sdpmDouble::getTape().isActive())
        preAccHlpr.finish(false);
}
inline void addPreAccIn(sdpmDouble const *v, size_t size = 1)
{
    if (sdpmDouble::getTape().isActive())
        for (size_t i = 0; i < size; ++i)
            preAccHlpr.addInput(v[i]);
}
inline void addPreAccOut(sdpmDouble *v, size_t size = 1)
{
    if (sdpmDouble::getTape().isActive())
        for (size_t i = 0; i < size; ++i)
            preAccHlpr.addOutput(v[i]);
}

// Matrix, vector and solver templates
template <typename T>
using Matrix = Eigen::Matrix<T, Eigen::Dynamic, Eigen::Dynamic>;
template <typename T>
using Vector = Eigen::Matrix<T, Eigen::Dynamic, 1>;
template <typename T>
inline void solve(Matrix<T> const &A, Vector<T> const &b, Vector<T> &x)
{
    x = A.partialPivLu().solve(b);
}

// CoDiPack <-> Eigen LinearSystem interface
template <typename T>
struct EigenSolver : public codi::EigenLinearSystem<T, Matrix, Vector>
{
public:
    using Base = codi::EigenLinearSystem<T, Matrix, Vector>;
    using MatrixReal = typename Base::MatrixReal;
    using VectorReal = typename Base::VectorReal;

    inline void solveSystem(MatrixReal const *A, VectorReal const *b, VectorReal *x)
    {
        solve(*A, *b, *x);
    }
};

// Solver helper function
inline sdpmVectorXd solve(sdpmMatrixXd &A, sdpmVectorXd &b)
{
    sdpmVectorXd x(b.size());
    codi::solveLinearSystem(EigenSolver<sdpmDouble>(), A, b, x);
    return x;
}

#endif // USE_CODI

} // namespace sdpm
#endif // SDPM_AD_INTERFACE_H
