# -*- coding: utf-8 -*-

# Copyright 2023 University of Liège
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# LANN wing (gradient)
# Adrien Crovato

import sdpm
from sdpm.gmsh import MeshLoader
from sdpm.utils import build_fpath
from sdpm.testing import *
import numpy as np

def main():
    # geometry parameters
    wnc = 20 # number of chordwise panels
    pars = {'nC': wnc, 'bC': 0.25, 'rS': 1}
    sref = 0.253 # reference area
    crd = 0.361 # root chord
    xf = 0.224 # x-coordinate of rotation center
    # flow and time parameters
    minf = 0.621 # freestream Mach number
    aoa = 0.59 * np.pi / 180
    kred = 0.133 # reduced frequency
    a_amp = 0.5 * 0.25 * np.pi / 180 # semi pitching amplitude
    h_amp = 0.1 # plunging amplitude

    # start timer
    tms = sdpm.Timers()
    tms['total'].start()
    # mesh
    tms['msh'].start()
    msh = MeshLoader(build_fpath('models/lann.geo', __file__, 1)).run(pars)
    tms['msh'].stop()
    # problem
    tms['pbl'].start()
    pbl = sdpm.Problem(msh)
    pbl.setGeometry(sref, crd, xf, 0., 0., True)
    pbl.setFreestream(aoa, 0., minf)
    pbl.setUnsteady(1, [kred])
    wing = sdpm.Body(msh, 'wing', 'te', crd, wnc, 1, 1)
    wing.addMotion()
    wing.setRigidMotion(0, a_amp, h_amp)
    pbl.addBody(wing)
    tms['pbl'].stop()
    # solver
    tms['sol'].start()
    sol = sdpm.Solver(pbl)
    sol.update()
    sol.run()
    sol.save(sdpm.GmshExport(msh))
    tms['sol'].stop()
    # gradient evaluation
    tms['grd'].start()
    grd = sdpm.Gradient(sol)
    grd.run()
    tms['grd'].stop()
    # end timer
    tms['total'].stop()
    print(tms)

    # recover GAF matrix from partial gradients (NB: since GAF is quadratic in DOF, result must be divided by 2)
    q_re = float(sol.getGafMatrixRe(0)[0][0])
    q_im = float(sol.getGafMatrixIm(0)[0][0])
    dq_re = 0.5 * (float(grd.computePartialsGafRe(0, 0, 0, 0, 'a')[0]) * a_amp + float(grd.computePartialsGafRe(0, 0, 0, 0, 'h')[0]) * h_amp)
    dq_im = 0.5 * (float(grd.computePartialsGafIm(0, 0, 0, 0, 'a')[0]) * a_amp + float(grd.computePartialsGafIm(0, 0, 0, 0, 'h')[0]) * h_amp)

    # test
    tests = CTests()
    tests.add(CTest('Re(Q)', q_re - dq_re, 0.0, 1e-10, forceabs=True))
    tests.add(CTest('Im(Q)', q_im - dq_im, 0.0, 1e-10, forceabs=True))
    tests.run()

    # eof
    print('')

if __name__ == '__main__':
    main()
