# -*- coding: utf-8 -*-

# Copyright 2023 University of Liège
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# NACA 0012 wing
# Adrien Crovato

import sdpm
from sdpm.gmsh import MeshLoader
from sdpm.utils import build_fpath
from sdpm.testing import *
import math

def main():
    # geometry parameters
    sym = 0
    span = 5 if sym==0 else 2.5 # span
    pars = {'symY': sym, 'wSpn': span, 'wNc': 100, 'wNs': 10}
    # flow parameters
    aoa = 3 * math.pi / 180

    # start timer
    tms = sdpm.Timers()
    tms['total'].start()
    # mesh
    tms['msh'].start()
    msh = MeshLoader(build_fpath('models/naca0012.geo', __file__, 1)).run(pars)
    tms['msh'].stop()
    # problem
    tms['pbl'].start()
    pbl = sdpm.Problem(msh)
    pbl.setGeometry(span, 1., 0., 0., 0., bool(sym))
    pbl.setFreestream(aoa, 0., 0.)
    wing = sdpm.Body(msh, 'wing', 'wTe', 1.)
    pbl.addBody(wing)
    tms['pbl'].stop()
    # solver
    tms['sol'].start()
    sol = sdpm.Solver(pbl)
    sol.update()
    sol.run()
    sol.save(sdpm.GmshExport(msh))
    tms['sol'].stop()
    # end timer
    tms['total'].stop()
    print(tms)

    # test
    tests = CTests()
    tests.add(CTest('CL', sol.getLiftCoeff(), 0.229, 5e-2)) # github.com/acrovato/aero: 0.226
    tests.add(CTest('CD', sol.getDragCoeff(), 0.003, 5e-2)) # aero: 0.003
    tests.add(CTest('CS', sol.getSideforceCoeff(), 0.000, 5e-2))
    tests.add(CTest('CM', sol.getMomentCoeff(), -0.056, 5e-2))
    tests.run()

    # eof
    print('')

if __name__ == '__main__':
    main()
