# -*- coding: utf-8 -*-

# Copyright 2023 University of Liège
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# NACA 0012 wing (adjoint)
# Adrien Crovato

import sdpm
from sdpm.gmsh import MeshLoader
from sdpm.utils import build_fpath
from sdpm.testing import *
import numpy as np

def main():
    # geometry parameters
    sym = 1
    span = 5 if sym==0 else 2.5 # span
    pars = {'symY': sym, 'wSpn': span, 'wNc': 75, 'wNs': 10}
    # flow parameters
    aoa = 3 * np.pi / 180
    minf = 0.5

    # start timer
    tms = sdpm.Timers()
    tms['total'].start()
    # mesh
    tms['msh'].start()
    msh = MeshLoader(build_fpath('models/naca0012.geo', __file__, 1)).run(pars)
    tms['msh'].stop()
    # problem
    tms['pbl'].start()
    pbl = sdpm.Problem(msh)
    pbl.setGeometry(span, 1., 0., 0., 0., bool(sym))
    pbl.setFreestream(aoa, 0., minf)
    wing = sdpm.Body(msh, 'wing', 'wTe', 1.)
    pbl.addBody(wing)
    tms['pbl'].stop()
    # solver and adjoint
    tms['sol'].start()
    sol = sdpm.Solver(pbl)
    adj = sdpm.Adjoint(sol)
    adj.solve()
    sol.save(sdpm.GmshExport(msh))
    tms['sol'].stop()
    # adjoint evaluation
    tms['adj'].start()
    d_cl = adj.compute('cl', [1.], ['aoa', 'x_wing'])
    d_cd = adj.compute('cd', [1.], ['aoa', 'x_wing'])
    tms['adj'].stop()
    # end timer
    tms['total'].stop()
    print(tms)

    # recover gradient wrt to AoA from gradient wrt mesh
    drot = np.array([[-np.sin(aoa), 0, np.cos(aoa)], [0, 0, 0], [-np.cos(aoa), 0, -np.sin(aoa)]])
    d_cl_aoa = 0
    d_cd_aoa = 0
    nodes = wing.getNodes()
    for i in range(len(nodes)):
        x = nodes[i].getCoords()
        dx = drot.dot(np.array([x[0] - 0.25, x[1], x[2]]))
        d_cl_aoa += np.array([d_cl['x_wing'][i * 3], d_cl['x_wing'][i * 3 + 1], d_cl['x_wing'][i * 3 + 2]]).dot(dx)
        d_cd_aoa += np.array([d_cd['x_wing'][i * 3], d_cd['x_wing'][i * 3 + 1], d_cd['x_wing'][i * 3 + 2]]).dot(dx)

    # test
    tests = CTests()
    tests.add(CTest('dCL/dAlpha', d_cl['aoa'][0], 4.58, 5e-2))
    tests.add(CTest('dCL/dAlpha (msh)', d_cl_aoa, d_cl['aoa'][0], 5e-2))
    tests.add(CTest('dCD/dAlpha', d_cd['aoa'][0], 0.127, 5e-2))
    tests.add(CTest('dCD/dAlpha (msh)', d_cd_aoa, d_cd['aoa'][0], 5e-2))
    tests.run()

    # eof
    print('')

if __name__ == '__main__':
    main()
